@@ -111,9 +111,9 @@ typedef struct
    * handshake, it becomes FALSE (and need_finish_handshake becomes
    * TRUE) at the end of the handshaking thread (and then the next
    * non-close op will call finish_handshake()). We can't just wait
-   * for handshake_thread_completed() to run, because it's possible
-   * that its main loop is being blocked by a synchronous op which is
-   * waiting for handshaking to become FALSE...
+   * for async_handshake_thread_completed() to run, because it's
+   * possible that its main loop is being blocked by a synchronous op
+   * which is waiting for handshaking to become FALSE...
    *
    * started_handshake indicates that the current handshake attempt
    * got at least as far as sending the first handshake packet (and so
@@ -1170,23 +1170,21 @@ verify_peer_certificate (GTlsConnectionBase *tls,
                          GTlsCertificate    *peer_certificate)
 {
   GTlsConnectionBaseClass *tls_class = G_TLS_CONNECTION_BASE_GET_CLASS (tls);
-  GSocketConnectable *peer_identity = NULL;
+  GSocketConnectable *peer_identity;
   GTlsDatabase *database;
-  GTlsCertificateFlags errors = 0;
+  GTlsCertificateFlags errors;
   gboolean is_client;
 
   is_client = G_IS_TLS_CLIENT_CONNECTION (tls);
 
-  if (is_client)
-    {
-      if (!g_tls_connection_base_is_dtls (tls))
-        peer_identity = g_tls_client_connection_get_server_identity (G_TLS_CLIENT_CONNECTION (tls));
-      else
-        peer_identity = g_dtls_client_connection_get_server_identity (G_DTLS_CLIENT_CONNECTION (tls));
+  if (!is_client)
+    peer_identity = NULL;
+  else if (!g_tls_connection_base_is_dtls (tls))
+    peer_identity = g_tls_client_connection_get_server_identity (G_TLS_CLIENT_CONNECTION (tls));
+  else
+    peer_identity = g_dtls_client_connection_get_server_identity (G_DTLS_CLIENT_CONNECTION (tls));
 
-      if (!peer_identity)
-        errors |= G_TLS_CERTIFICATE_BAD_IDENTITY;
-    }
+  errors = 0;
 
   database = g_tls_connection_get_database (G_TLS_CONNECTION (tls));
   if (!database)
@@ -1308,6 +1306,9 @@ accept_or_reject_peer_certificate (gpointer user_data)
   g_cond_signal (&priv->verify_certificate_condition);
   g_mutex_unlock (&priv->verify_certificate_mutex);
 
+  g_object_notify (G_OBJECT (tls), "peer-certificate");
+  g_object_notify (G_OBJECT (tls), "peer-certificate-errors");
+
   return G_SOURCE_REMOVE;
 }
 
@@ -1453,16 +1454,10 @@ crank_sync_handshake_context (GTlsConnectionBase *tls,
   /* need_finish_handshake will be set inside sync_handshake_thread_completed(),
    * which should only ever be invoked while iterating the handshake context
    * here. So need_finish_handshake should only change on this thread.
-   *
-   * FIXME: This function is not cancellable. We should figure out how to
-   * support cancellation. We must not return from this function before it is
-   * safe to destroy handshake_context, but it's not safe to destroy
-   * handshake_context until after the handshake has completed. And the
-   * handshake operation is not cancellable, so we have a problem.
    */
   g_mutex_lock (&priv->op_mutex);
   priv->sync_handshake_in_progress = TRUE;
-  while (priv->sync_handshake_in_progress)
+  while (priv->sync_handshake_in_progress && !g_cancellable_is_cancelled (cancellable))
     {
       g_mutex_unlock (&priv->op_mutex);
       g_main_context_iteration (priv->handshake_context, TRUE);
@@ -1584,15 +1579,15 @@ g_tls_connection_base_dtls_handshake (GDtlsConnection  *conn,
 }
 
 /* In the async version we use two GTasks; one to run
- * handshake_thread() and then call handshake_thread_completed(), and
- * a second to call the caller's original callback after we call
+ * handshake_thread() and then call async_handshake_thread_completed(),
+ * and a second to call the caller's original callback after we call
  * finish_handshake().
  */
 
 static void
-handshake_thread_completed (GObject      *object,
-                            GAsyncResult *result,
-                            gpointer      user_data)
+async_handshake_thread_completed (GObject      *object,
+                                  GAsyncResult *result,
+                                  gpointer      user_data)
 {
   GTask *caller_task = user_data;
   GTlsConnectionBase *tls = g_task_get_source_object (caller_task);
@@ -1613,33 +1608,20 @@ handshake_thread_completed (GObject      *object,
     need_finish_handshake = FALSE;
   g_mutex_unlock (&priv->op_mutex);
 
-  /* We have to clear handshake_context before g_task_return_* because it can
-   * return immediately to application code inside g_task_return_*,
-   * and the application code could then start a new TLS operation.
-   *
-   * But we can't clear until after finish_handshake().
-   */
   if (need_finish_handshake)
     {
       success = finish_handshake (tls, G_TASK (result), &error);
-
-      g_clear_pointer (&priv->handshake_context, g_main_context_unref);
-
       if (success)
         g_task_return_boolean (caller_task, TRUE);
       else
         g_task_return_error (caller_task, error);
     }
+  else if (priv->handshake_error)
+    g_task_return_error (caller_task, g_error_copy (priv->handshake_error));
   else
-    {
-      g_clear_pointer (&priv->handshake_context, g_main_context_unref);
-
-      if (priv->handshake_error)
-        g_task_return_error (caller_task, g_error_copy (priv->handshake_error));
-      else
-        g_task_return_boolean (caller_task, TRUE);
-    }
+    g_task_return_boolean (caller_task, TRUE);
 
+  g_clear_pointer (&priv->handshake_context, g_main_context_unref);
   g_object_unref (caller_task);
 }
 
@@ -1691,7 +1673,7 @@ g_tls_connection_base_handshake_async (GTlsConnection      *conn,
   g_task_set_name (caller_task, "[glib-networking] g_tls_connection_base_handshake_async (caller task)");
   g_task_set_priority (caller_task, io_priority);
 
-  thread_task = g_task_new (conn, cancellable, handshake_thread_completed, caller_task);
+  thread_task = g_task_new (conn, cancellable, async_handshake_thread_completed, caller_task);
   g_task_set_source_tag (thread_task, g_tls_connection_base_handshake_async);
   g_task_set_name (caller_task, "[glib-networking] g_tls_connection_base_handshake_async (thread task)");
   g_task_set_priority (thread_task, io_priority);
@@ -2433,12 +2415,10 @@ g_tls_connection_base_get_base_ostream (GTlsConnectionBase *tls)
 }
 
 void
-g_tls_connection_base_set_missing_requested_client_certificate (GTlsConnectionBase *tls)
+g_tls_connection_base_handshake_thread_set_missing_requested_client_certificate (GTlsConnectionBase *tls)
 {
   GTlsConnectionBasePrivate *priv = g_tls_connection_base_get_instance_private (tls);
 
-  /* FIXME: Assert this is only used on the handshake thread. */
-
   priv->missing_requested_client_certificate = TRUE;
 }
 
@@ -2507,7 +2487,7 @@ g_tls_connection_base_ever_handshaked (GTlsConnectionBase *tls)
 }
 
 gboolean
-g_tls_connection_base_request_certificate (GTlsConnectionBase *tls)
+g_tls_connection_base_handshake_thread_request_certificate (GTlsConnectionBase *tls)
 {
   GTlsConnectionBasePrivate *priv = g_tls_connection_base_get_instance_private (tls);
   GTlsInteractionResult res = G_TLS_INTERACTION_UNHANDLED;
@@ -2611,11 +2591,11 @@ GTLS_DEBUG (gpointer    connection,
             ...)
 {
   char *result = NULL;
+  va_list args;
   int ret;
 
   g_assert (G_IS_TLS_CONNECTION (connection));
 
-  va_list args;
   va_start (args, message);
 
   ret = g_vasprintf (&result, message, args);
